# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2007, Agendaless Consulting and Contributors.
# Copyright (c) 2008, Florent Aide <florent.aide@gmail.com> and
#                     Gustavo Narea <me@gustavonarea.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Sample plugins and middleware configuration for repoze.what."""

from repoze.who.plugins.auth_tkt import AuthTktCookiePlugin
from repoze.who.plugins.sa import SQLAlchemyAuthenticatorPlugin

from repoze.what.middleware import setup_auth
from repoze.what.plugins.sql import configure_sql_adapters


def find_plugin_translations(translations={}):
    """
    Process translations defined in TG2 quickstarted projects.
    
    These "translations" are usually defined in quickstarted projects, in
    {tg2_project}.config.app_cfg.base_config.sa_auth.translations
    
    @param translations: The TG2 applications' base_config.sa_auth.translations
    @return: The respective translations for the group and permission adapters
        and the authenticator.
    @rtype: C{dict}
    
    """
    
    group_adapter = {}
    permission_adapter = {}
    authenticator = {}
    
    if 'validate_password' in translations:
        authenticator['validate_password'] = translations['validate_password']
    if 'user_name' in translations:
        group_adapter['item_name'] = translations['user_name']
        authenticator['user_name'] = translations['user_name']
    if 'users' in translations:
        group_adapter['items'] = translations['users']
    if 'group_name' in translations:
        group_adapter['section_name'] = translations['group_name']
        permission_adapter['item_name'] = translations['group_name']
    if 'groups' in translations:
        group_adapter['sections'] = translations['groups']
        permission_adapter['items'] = translations['groups']
    if 'permission_name' in translations:
        permission_adapter['section_name'] = translations['permission_name']
    if 'permissions' in translations:
        permission_adapter['sections'] = translations['permissions']
        
    final_translations = {
        'group_adapter': group_adapter,
        'permission_adapter': permission_adapter,
        'authenticator': authenticator}
    return final_translations


def setup_sql_auth(app, user_class, group_class, permission_class,
                   dbsession, form_plugin=None, form_identifies=True,
                   cookie_secret='secret', cookie_name='authtkt',
                   translations={}, **who_args):
    """
    Configure :mod:`repoze.who` and :mod:`repoze.what` with SQL-only 
    authentication and authorization, respectively.
    
    :param app: Your WSGI application.
    :param user_class: The SQLAlchemy/Elixir class for the users.
    :param group_class: The SQLAlchemy/Elixir class for the groups.
    :param permission_class: The SQLAlchemy/Elixir class for the permissions.
    :param dbsession: The SQLAlchemy/Elixir session.
    :param form_plugin: The main :mod:`repoze.who` challenger plugin; this is 
        usually a login form.
    :param form_identifies: Whether the ``form_plugin`` may and should act as
        an :mod:`repoze.who` identifier.
    :type form_identifies: bool
    :param cookie_secret: The "secret" for the AuthTktCookiePlugin.
    :type cookie_secret: str
    :param cookie_name: The name for the AuthTktCookiePlugin.
    :type cookie_name: str
    :param translations: The model translations.
    :return: The WSGI application with authentication and authorization
        middleware.
    
    It configures :mod:`repoze.who` with the following plugins:
    
    * Identifiers:
    
      * :class:`repoze.who.plugins.auth_tkt.AuthTktCookiePlugin`. You can
        customize the cookie name and secret using the ``cookie_name`` and
        ``cookie_secret`` arguments, respectively.
      * :class:`repoze.who.plugins.form.RedirectingFormPlugin` as the first
        identifier and challenger -- using ``/login`` as the relative URL that 
        will \display the login form, ``/login_handler`` as the relative URL 
        where the form will be sent and ``/logout_handler`` as the relative URL
        where the user will be logged out. The so-called *rememberer* of such 
        identifier will be an instance of 
        :class:`repoze.who.plugins.auth_tkt.AuthTktCookiePlugin`.
        
        .. tip::
        
            This plugin may be overriden with the ``form_plugin`` argument. See
            also the ``form_identifies`` argument.
      
      Then it will append the identifiers you pass through the ``identifiers``
      keyword argument, if any.
    
    * Authenticators:
    
      * :class:`repoze.who.plugins.sa.SQLAlchemyAuthenticatorPlugin`, using
        the ``user_class`` and ``dbsession`` arguments as its user class and
        DB session, respectively.
      
      Then it will append the authenticators you pass through the 
      ``authenticators`` keyword argument, if any.
    
    * Challengers:
    
      * The same Form-based plugin used in the identifiers.
      
      Then it will append the challengers you pass through the 
      ``challengers`` keyword argument, if any.
    
    * Metadata providers: The metadata providers you pass through the 
      ``mdproviders`` keyword argument, if any.
    
    Additional keyword arguments will be passed to 
    :class:`repoze.who.middleware.PluggableAuthenticationMiddleware`.
    
    """
    plugin_translations = find_plugin_translations(translations)
    
    source_adapters = configure_sql_adapters(
        user_class,
        group_class,
        permission_class,
        dbsession,
        plugin_translations['group_adapter'],
        plugin_translations['permission_adapter']
        )
    group_adapters = {'sql_auth': source_adapters['group']}
    permission_adapters = {'sql_auth': source_adapters['permission']}
    
    # Setting the repoze.who authenticators:
    sqlauth = SQLAlchemyAuthenticatorPlugin(user_class, dbsession)
    sqlauth.translations.update(plugin_translations['authenticator'])
    if 'authenticators' not in who_args:
        who_args['authenticators'] = []
    who_args['authenticators'].append(('sqlauth', sqlauth))
    
    cookie = AuthTktCookiePlugin(cookie_secret, cookie_name)
    
    # Setting the repoze.who identifiers
    if 'identifiers' not in who_args:
        who_args['identifiers'] = []
    who_args['identifiers'].append(('cookie', cookie))
    
    if form_plugin is None:
        from repoze.who.plugins.form import RedirectingFormPlugin
        form = RedirectingFormPlugin('/login', '/login_handler',
                                     '/logout_handler',
                                     rememberer_name='cookie')
    else:
        form = form_plugin
    
    if form_identifies:
        who_args['identifiers'].insert(0, ('main_identifier', form))
    
    # Setting the repoze.who challengers:
    if 'challengers' not in who_args:
        who_args['challengers'] = []
    who_args['challengers'].append(('form', form))
    
    middleware = setup_auth(app, group_adapters, permission_adapters, 
                            **who_args)
    return middleware



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, Gustavo Narea <me@gustavonarea.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE
#
##############################################################################

"""
Tests for the repoze.what SQL quickstart.

"""

import unittest

from repoze.who.plugins.basicauth import BasicAuthPlugin
from repoze.who.plugins.auth_tkt import AuthTktCookiePlugin
from repoze.who.plugins.form import RedirectingFormPlugin
from repoze.who.plugins.sa import SQLAlchemyAuthenticatorPlugin
from repoze.who.tests import Base as BasePluginTester, DummyApp

from repoze.what.middleware import AuthorizationMetadata
from repoze.what.plugins.quickstart import setup_sql_auth

import databasesetup
from fixture.model import User, Group, Permission, DBSession


class TestSetupAuth(BasePluginTester):
    """Tests for the setup_sql_auth() function"""
    
    def setUp(self):
        super(TestSetupAuth, self).setUp()
        databasesetup.setup_database()
    
    def tearDown(self):
        super(TestSetupAuth, self).tearDown()
        databasesetup.teardownDatabase()
    
    def _in_registry(self, app, registry_key, registry_type):
        assert registry_key in app.name_registry, ('Key "%s" not in registry' %
                                                   registry_key)
        assert isinstance(app.name_registry[registry_key], registry_type), \
               'Registry key "%s" is of type "%s" (not "%s")' % \
               (registry_key, app.name_registry[registry_key].__class__.__name__,
                registry_type.__name__)
    
    def _makeApp(self, **who_args):
        app_with_auth = setup_sql_auth(DummyApp(), User, Group, Permission,
                                       DBSession, **who_args)
        return app_with_auth

    def test_no_extras(self):
        app = self._makeApp()
        self._in_registry(app, 'main_identifier', RedirectingFormPlugin)
        self._in_registry(app, 'authorization_md', AuthorizationMetadata)
        self._in_registry(app, 'cookie', AuthTktCookiePlugin)
        self._in_registry(app, 'sqlauth', SQLAlchemyAuthenticatorPlugin)
        self._in_registry(app, 'form', RedirectingFormPlugin)
    
    def test_form_doesnt_identify(self):
        app = self._makeApp(form_identifies=False)
        assert 'main_identifier' not in app.name_registry
    
    def test_additional_identifiers(self):
        identifiers = [('http_auth', BasicAuthPlugin('1+1=2'))]
        app = self._makeApp(identifiers=identifiers)
        self._in_registry(app, 'main_identifier', RedirectingFormPlugin)
        self._in_registry(app, 'http_auth', BasicAuthPlugin)
    
    def test_non_default_form_plugin(self):
        app = self._makeApp(form_plugin=BasicAuthPlugin('1+1=2'))
        self._in_registry(app, 'main_identifier', BasicAuthPlugin)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, Gustavo Narea <me@gustavonarea.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE
#
##############################################################################

"""
SQL plugin for repoze.what based on SQLAlchemy.

This is a core plugin that provides ``source adapters`` for groups and
permissions stored in databases.

Below is the "translation" of the terminology used by repoze.what into
databases' terminology:
 * source -> database
 * section -> parent row (in a one-to-many relationship)
 * item -> a child row (in a one-to-many relationship)

In a group source adapter, a ``section`` is a group in the source (a parent
row) and its items are the users that belong to that section (its children
rows). In a permission source adapter, a ``section`` is a permission in the
source (a parent row) and its items are the groups that are granted that
permission (its children rows).

For developers to be able to use the names they want in their model, both the
groups and permissions source adapters use a "translation table" for the
field and table names involved:
  * Group source adapter:
    * "section_name" (default: "group_name"): The name of the table field that
      contains the primary key in the groups table.
    * "sections" (default: "groups"): The groups to which a given user belongs.
    * "item_name" (default: "user_name"): The name of the table field that
      contains the primary key in the users table.
    * "items" (default: "users"): The users that belong to a given group.
  * Permission source adapter:
    * "section_name" (default: "permission_name"): The name of the table field
      that contains the primary key in the permissions table.
    * "sections" (default: "permissions"): The permissions granted to a given
      group.
    * "item_name" (default: "group_name"): The name of the table field that
      contains the primary key in the groups table.
    * "items" (default: "groups"): The groups that are granted a given
      permission.

@attention: These adapters have a serious limitation, since they are based on
    one-to-many relations: If you're using the groups adapter, then your users'
    data must also be handled by SQLAlchemy; likewise, If you're using the
    permissions adapter, then your groups must also be handled by SQLAlchemy.
@attention: When using the SQL group adapter, it will load the authenticated
    used object into repoze.who's identity dict (under the "user" key).

"""

from sqlalchemy.exceptions import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound

from repoze.what.adapters import BaseSourceAdapter, SourceError

__all__ = ['SqlGroupsAdapter', 'SqlPermissionsAdapter',
           'configure_sql_adapters']


class _BaseSqlAdapter(BaseSourceAdapter):
    """Base class for SQL source adapters."""

    def __init__(self, parent_class, children_class, dbsession):
        """
        Create an SQL source adapter.

        @param parent_class: The SQLAlchemy table of the section.
        @param children_class: The SQLAlchemy table of the items.
        @param dbsession: The SQLAlchemy session.

        """
        super(_BaseSqlAdapter, self).__init__()
        self.dbsession = dbsession
        self.parent_class = parent_class
        self.children_class = children_class

    # BaseSourceAdapter
    def _get_all_sections(self):
        sections = {}
        sections_as_rows = self.dbsession.query(self.parent_class).all()
        for section_as_row in sections_as_rows:
            section_name = getattr(section_as_row,
                                   self.translations['section_name'])
            sections[section_name] = self._get_section_items(section_name)
        return sections

    # BaseSourceAdapter
    def _get_section_items(self, section):
        section_as_row = self._get_section_as_row(section)
        # A short-cut a translation:
        item_name = self.translations['item_name']
        # The name of all the items that belong to the section in question:
        items_as_rowset = getattr(section_as_row, self.translations['items'])
        return set((getattr(i, item_name) for i in items_as_rowset))

    # BaseSourceAdapter
    def _find_sections(self, hint):
        raise NotImplementedError('This is implemented in the groups and '
                                  'permissions adapters.')

    # TODO: Factor out the common code in _include_items and _exclude_items

    # BaseSourceAdapter
    def _include_items(self, section, items):
        self.dbsession.begin(subtransactions=True)
        item, included_items = self._get_items_as_rowset(section)
        try:
            for item_to_include in items:
                item_as_row = self._get_item_as_row(item_to_include)
                included_items.append(item_as_row)
            self.dbsession.commit()
        except SQLAlchemyError, msg:
            self.dbsession.rollback()
            raise SourceError(msg)

    # BaseSourceAdapter
    def _exclude_items(self, section, items):
        self.dbsession.begin(subtransactions=True)
        item, included_items = self._get_items_as_rowset(section)
        try:
            for item_to_exclude in items:
                item_as_row = self._get_item_as_row(item_to_exclude)
                included_items.remove(item_as_row)
            self.dbsession.commit()
        except SQLAlchemyError, msg:
            self.dbsession.rollback()
            raise SourceError(msg)

    # BaseSourceAdapter
    def _item_is_included(self, section, item):
        return item in self._get_section_items(section)

    # BaseSourceAdapter
    def _create_section(self, section):
        self.dbsession.begin(subtransactions=True)
        section_as_row = self.parent_class()
        # Creating the section with an empty set of items:
        setattr(section_as_row, self.translations['section_name'], section)
        setattr(section_as_row, self.translations['items'], [])
        try:
            self.dbsession.add(section_as_row)
            self.dbsession.commit()
        except SQLAlchemyError, msg:
            self.dbsession.rollback()
            raise SourceError(msg)

    # BaseSourceAdapter
    def _edit_section(self, section, new_section):
        self.dbsession.begin(subtransactions=True)
        section_as_row = self._get_section_as_row(section)
        setattr(section_as_row, self.translations['section_name'], new_section)
        try:
            self.dbsession.commit()
        except SQLAlchemyError, msg:
            self.dbsession.rollback()
            raise SourceError(msg)

    # BaseSourceAdapter
    def _delete_section(self, section):
        self.dbsession.begin(subtransactions=True)
        section_as_row = self._get_section_as_row(section)
        try:
            self.dbsession.delete(section_as_row)
            self.dbsession.commit()
        except SQLAlchemyError, msg:
            self.dbsession.rollback()
            raise SourceError(msg)

    # BaseSourceAdapter
    def _section_exists(self, section):
        # TODO: There must be a more elegant way to do this with SQLAlchemy
        try:
            self._get_section_as_row(section)
            return True
        except SourceError:
            return False

    def _get_section_as_row(self, section_name):
        """
        Return the SQLAlchemy row for the section called C{section_name}.

        When dealing with a group source, the section is a group. And when
        dealing with a permission source, the section is a permission.

        """
        # "field" usually equals to {tg_package}.model.Group.group_name
        # or {tg_package}.model.Permission.permission_name
        field = getattr(self.parent_class, self.translations['section_name'])
        query = self.dbsession.query(self.parent_class)
        try:
            section_as_row = query.filter(field==section_name).one()
        except NoResultFound:
            msg = 'Section (%s) "%s" is not defined in the parent table'
            msg = msg % (self.translations['section_name'], section_name)
            raise SourceError(msg)
        return section_as_row

    def _get_item_as_row(self, item_name):
        """
        Return the SQLAlchemy row for the item called C{item_name}.

        When dealing with a group source, the item is a user. And when dealing
        with a permission source, the item is a group.

        """
        # "field" usually equals to {tg_package}.model.User.user_name
        # or {tg_package}.model.Group.group_name
        field = getattr(self.children_class, self.translations['item_name'])
        query = self.dbsession.query(self.children_class)
        try:
            item_as_row = query.filter(field==item_name).one()
        except NoResultFound:
            msg = 'Item (%s) "%s" does not exist in the child table'
            msg = msg % (self.translations['item_name'], item_name)
            raise SourceError(msg)
        return item_as_row

    def _get_items_as_rowset(self, section_name):
        """
        Return the items of the section called C{section_name}.

        When dealing with a group source, the section is a group and the
        items (the result of this function) are the users that belong to such
        a group.

        When dealing with a permission source, the section is a permission and
        the items (the result of this function) are the groups that are granted
        such a permission.

        """
        section_as_row = self._get_section_as_row(section_name)
        items_as_rowset = getattr(section_as_row, self.translations['items'])
        return section_as_row, items_as_rowset


#{ Source adapters


class SqlGroupsAdapter(_BaseSqlAdapter):
    """
    The SQL group source adapter.
    
    To use this adapter, you must also define your users in a SQLAlchemy or
    Elixir-managed table with the relevant one-to-many (or many-to-many) 
    relationship defined with ``group_class``.
    
    On the other hand, unless stated otherwise, it will also assume the 
    following naming conventions in both classes; to replace any of those
    default values, you should use the ``translations`` dictionary of the
    relevant class accordingly:
    
    * In `group_class`, the attribute that contains the group name is 
      ``group_name`` (e.g., ``Group.group_name``).
    * In `group_class`, the attribute that contains the members of such a group
      is ``users`` (e.g., ``Group.users``).
    * In `user_class`, the attribute that contains the user's name is
      ``user_name`` (e.g., ``User.user_name``).
    * In `user_class`, the attribute that contains the groups to which a user
      belongs is ``groups`` (e.g., ``User.groups``).
    
    Example #1, without special naming conventions::
    
        # ...
        from repoze.what.plugins.sql import SqlGroupsAdapter
        from my_model import User, Group, DBSession
        
        groups = SqlGroupsAdapter(Group, User, DBSession)
        
        # ...
    
    Example #2, with special naming conventions::
    
        # ...
        from repoze.what.plugins.sql import SqlGroupsAdapter
        from my_model import Member, Team, DBSession
        
        groups = SqlGroupsAdapter(Team, Member, DBSession)
        
        # Replacing the default attributes, if necessary:
        
        # We have "Team.team_name" instead of "Team.group_name":
        groups.translations['section_name'] = 'team_name'
        # We have "Team.members" instead of "Team.users":
        groups.translations['items'] = 'members'
        # We have "Member.username" instead of "Member.user_name":
        groups.translations['item_name'] = 'username'
        # We have "Member.teams" instead of "Member.groups":
        groups.translations['sections'] = 'teams'
        
        # ...
    
    """

    def __init__(self, group_class, user_class, dbsession):
        """
        Create an SQL groups source adapter.
    
        :param group_class: The class that manages the groups.
        :param user_class: The class that manages the users.
        :param dbsession: The SQLALchemy/Elixir session to be used.
        
        """
        super(SqlGroupsAdapter, self).__init__(parent_class=group_class,
                                               children_class=user_class,
                                               dbsession=dbsession)
        self.translations = {
            'section_name': 'group_name',
            'sections': 'groups',
            'item_name': 'user_name',
            'items': 'users'
        }

    def _find_sections(self, identity):
        id_ = identity['repoze.who.userid']
        try:
            user = self._get_item_as_row(id_)
        except SourceError:
            return set()

        # Also load the user object into the identity
        identity['user'] = user

        user_memberships = getattr(user, self.translations['sections'])
        return set([getattr(group, self.translations['section_name'])
                    for group in user_memberships])


class SqlPermissionsAdapter(_BaseSqlAdapter):
    """
    The SQL permission source adapter.
    
    To use this adapter, you must also define your groups in a SQLAlchemy or
    Elixir-managed table with the relevant one-to-many (or many-to-many)
    relationship defined with ``permission_class``.
    
    On the other hand, unless stated otherwise, it will also assume the 
    following naming conventions in both classes; to replace any of those
    default values, you should use the ``translations`` dictionary of the
    relevant class accordingly:
    
    * In `permission_class`, the attribute that contains the permission name is 
      ``permission_name`` (e.g., ``Permission.permission_name``).
    * In `permission_class`, the attribute that contains the groups that are 
      granted such a permission is ``groups`` (e.g., ``Permission.groups``).
    * In `group_class`, the attribute that contains the group name is
      ``group_name`` (e.g., ``Group.group_name``).
    * In `group_class`, the attribute that contains the permissions granted to
      that group is ``permissions`` (e.g., ``Group.permissions``).
    
    Example #1, without special naming conventions::
    
        # ...
        from repoze.what.plugins.sql import SqlPermissionsAdapter
        from my_model import Group, Permission, DBSession
        
        groups = SqlPermissionsAdapter(Permission, Group, DBSession)
        
        # ...
    
    Example #2, with special naming conventions::
    
        # ...
        from repoze.what.plugins.sql import SqlPermissionsAdapter
        from my_model import Team, Permission, DBSession
        
        permissions = SqlPermissionsAdapter(Permission, Team, DBSession)
        
        # Replacing the default attributes, if necessary:
        
        # We have "Permission.perm_name" instead of "Permission.permission_name":
        permissions.translations['section_name'] = 'perm_name'
        # We have "Permission.teams" instead of "Permission.groups":
        permissions.translations['items'] = 'teams'
        # We have "Team.team_name" instead of "Team.group_name":
        permissions.translations['item_name'] = 'team_name'
        # We have "Team.perms" instead of "Team.permissions":
        permissions.translations['sections'] = 'perms'
        
        # ...
    
    
    """

    def __init__(self, permission_class, group_class, dbsession):
        """
        Create an SQL permissions source adapter.
        
        :param permission_class: The class that manages the permissions.
        :param group_class: The class that manages the groups.
        :param dbsession: The SQLALchemy/Elixir session to be used.
        
        """
        
        super(SqlPermissionsAdapter, self).__init__(
            parent_class=permission_class,
            children_class=group_class,
            dbsession=dbsession
            )
        self.translations = {
            'section_name': 'permission_name',
            'sections': 'permissions',
            'item_name': 'group_name',
            'items': 'groups'
        }

    def _find_sections(self, group_name):
        try:
            group = self._get_item_as_row(group_name)
        except SourceError:
            return set()

        group_permissions = getattr(group, self.translations['sections'])

        return set([getattr(permission, self.translations['section_name'])
                    for permission in group_permissions])


#{ Utilities


def configure_sql_adapters(user_class, group_class, permission_class, session,
                           group_translations={}, permission_translations={}):
    """
    Configure and return group and permission adapters that share the same model.
    
    :param user_class: The class that manages the users.
    :param group_class: The class that manages the groups.
    :param user_class: The class that manages the permissions.
    :param dbsession: The SQLALchemy/Elixir session to be used.
    :param group_translations: The dictionary of translations for the group.
    :param permission_translations: The dictionary of translations for the permissions.
    :return: The ``group`` and ``permission`` adapters, configured.
    :rtype: dict 
    
    For this function to work, ``user_class`` and ``group_class`` must have the
    relevant one-to-many (or many-to-many) relationship; likewise, 
    ``group_class`` and ``permission_class`` must have the relevant one-to-many 
    (or many-to-many) relationship.
    
    Example::
    
        # ...
        from repoze.what.plugins.sql import configure_sql_adapters
        from my_model import User, Group, Permission, DBSession
        
        adapters = configure_sql_adapters(User, Group, Permission, DBSession)
        groups = adapters['group']
        permissions = adapters['permission']
        
        # ...

    """
    # Creating the adapters:
    group = SqlGroupsAdapter(group_class, user_class, session)
    permission = SqlPermissionsAdapter(permission_class, group_class, session)
    # Translating some fields:
    group.translations.update(group_translations)
    permission.translations.update(permission_translations)

    return {'group': group, 'permission': permission}


#}
